@@ -1,4 +1,5 @@
 import datetime
+import logging
 
 import hiro
 from flask import Blueprint, Flask, current_app
@@ -493,8 +494,9 @@ def test_register_blueprint(extension_factory):
 
 
 def test_invalid_decorated_static_limit_blueprint(caplog):
+    caplog.set_level(logging.INFO)
     app = Flask(__name__)
-    limiter = Limiter(app, default_limits=["1/second"], key_func=get_remote_address)
+    limiter = Limiter(get_remote_address, app=app, default_limits=["1/second"])
     bp = Blueprint("bp1", __name__)
 
     @bp.route("/t1")
@@ -508,14 +510,15 @@ def test_invalid_decorated_static_limit_blueprint(caplog):
         with hiro.Timeline().freeze():
             assert cli.get("/t1").status_code == 200
             assert cli.get("/t1").status_code == 429
-    assert "failed to configure" in caplog.records[0].msg
-    assert "exceeded at endpoint" in caplog.records[1].msg
+    assert "failed to load" in caplog.records[0].msg
+    assert "exceeded at endpoint" in caplog.records[-1].msg
 
 
 def test_invalid_decorated_dynamic_limits_blueprint(caplog):
+    caplog.set_level(logging.INFO)
     app = Flask(__name__)
     app.config.setdefault("X", "2 per sec")
-    limiter = Limiter(app, default_limits=["1/second"], key_func=get_remote_address)
+    limiter = Limiter(get_remote_address, app=app, default_limits=["1/second"])
     bp = Blueprint("bp1", __name__)
 
     @bp.route("/t1")
