@@ -1,4 +1,5 @@
 import asyncio
+import logging
 from functools import wraps
 from unittest import mock
 
@@ -7,6 +8,7 @@ from flask import Blueprint, Flask, current_app, g, make_response, request
 from werkzeug.exceptions import BadRequest
 
 from flask_limiter import ExemptionScope, Limiter
+from flask_limiter.util import get_remote_address
 
 
 def get_ip_from_header():
@@ -18,7 +20,7 @@ def test_multiple_decorators(extension_factory):
 
     @app.route("/t1")
     @limiter.limit(
-        "100 per minute", lambda: "test"
+        "100 per minute", key_func=lambda: "test"
     )  # effectively becomes a limit for all users
     @limiter.limit("50/minute")  # per ip as per default key_func
     def t1():
@@ -73,6 +75,22 @@ def test_exempt_routes(extension_factory):
         assert cli.get("/t4").status_code == 429
 
 
+def test_decorated_limit_with_scope(extension_factory):
+    app, limiter = extension_factory()
+
+    @app.route("/t/<path:path>")
+    @limiter.limit("1/second", scope=lambda _: request.view_args["path"])
+    def t(path):
+        return "test"
+
+    with hiro.Timeline():
+        with app.test_client() as cli:
+            assert cli.get("/t/1").status_code == 200
+            assert cli.get("/t/1").status_code == 429
+            assert cli.get("/t/2").status_code == 200
+            assert cli.get("/t/2").status_code == 429
+
+
 def test_decorated_limit_with_conditional_deduction(extension_factory):
     app, limiter = extension_factory()
 
@@ -292,9 +310,10 @@ def test_decorated_dynamic_limits(extension_factory):
 
 
 def test_invalid_decorated_dynamic_limits(caplog):
+    caplog.set_level(logging.INFO)
     app = Flask(__name__)
     app.config.setdefault("X", "2 per sec")
-    limiter = Limiter(app, default_limits=["1/second"], key_func=get_ip_from_header)
+    limiter = Limiter(get_ip_from_header, app=app, default_limits=["1/second"])
 
     @app.route("/t1")
     @limiter.limit(lambda: current_app.config.get("X"))
@@ -310,12 +329,30 @@ def test_invalid_decorated_dynamic_limits(caplog):
     assert "failed to load ratelimit" in caplog.records[0].msg
     assert "failed to load ratelimit" in caplog.records[1].msg
     assert "exceeded at endpoint" in caplog.records[2].msg
-    assert caplog.records[2].levelname == "WARNING"
+    assert caplog.records[2].levelname == "INFO"
+
+
+def test_decorated_limit_empty_exempt(caplog):
+    app = Flask(__name__)
+    limiter = Limiter(get_remote_address, app=app)
+
+    @app.route("/t1")
+    @limiter.limit(lambda: "")
+    def t1():
+        return "42"
+
+    with app.test_client() as cli:
+        with hiro.Timeline().freeze():
+            assert cli.get("/t1").status_code == 200
+            assert cli.get("/t1").status_code == 200
+
+    assert not caplog.records
 
 
 def test_invalid_decorated_static_limits(caplog):
+    caplog.set_level(logging.INFO)
     app = Flask(__name__)
-    limiter = Limiter(app, default_limits=["1/second"], key_func=get_ip_from_header)
+    limiter = Limiter(get_ip_from_header, app=app, default_limits=["1/second"])
 
     @app.route("/t1")
     @limiter.limit("2/sec")
@@ -326,8 +363,8 @@ def test_invalid_decorated_static_limits(caplog):
         with hiro.Timeline().freeze():
             assert cli.get("/t1").status_code == 200
             assert cli.get("/t1").status_code == 429
-    assert "failed to configure" in caplog.records[0].msg
-    assert "exceeded at endpoint" in caplog.records[1].msg
+    assert "failed to load" in caplog.records[0].msg
+    assert "exceeded at endpoint" in caplog.records[-1].msg
 
 
 def test_named_shared_limit(extension_factory):
@@ -397,7 +434,7 @@ def test_dynamic_shared_limit(extension_factory):
 def test_conditional_limits():
     """Test that the conditional activation of the limits work."""
     app = Flask(__name__)
-    limiter = Limiter(app, key_func=get_ip_from_header)
+    limiter = Limiter(get_ip_from_header, app=app)
 
     @app.route("/limited")
     @limiter.limit("1 per day")
@@ -434,7 +471,7 @@ def test_conditional_limits():
 def test_conditional_shared_limits():
     """Test that conditional shared limits work."""
     app = Flask(__name__)
-    limiter = Limiter(app, key_func=get_ip_from_header)
+    limiter = Limiter(get_ip_from_header, app=app)
 
     @app.route("/limited")
     @limiter.shared_limit("1 per day", "test_scope")
@@ -468,13 +505,12 @@ def test_conditional_shared_limits():
 
 
 def test_whitelisting():
-
     app = Flask(__name__)
     limiter = Limiter(
-        app,
+        get_ip_from_header,
+        app=app,
         default_limits=["1/minute"],
         headers_enabled=True,
-        key_func=get_ip_from_header,
     )
 
     @app.route("/")
@@ -558,7 +594,6 @@ def test_decorated_limit_immediate(extension_factory):
 
 
 def test_decorated_shared_limit_immediate(extension_factory):
-
     app, limiter = extension_factory(default_limits=["1/minute"])
     shared = limiter.shared_limit(lambda: g.rate_limit, "shared")
 
@@ -615,8 +650,8 @@ def test_async_route(extension_factory):
         assert cli.get("/t2").status_code == 200
 
 
-def test_on_breach_callback(extension_factory, caplog):
-    app, limiter = extension_factory()
+def test_on_breach_callback_swallow_errors(extension_factory, caplog):
+    app, limiter = extension_factory(swallow_errors=True)
 
     callbacks = []
 
@@ -650,10 +685,10 @@ def test_on_breach_callback(extension_factory, caplog):
         assert cli.get("/fail").status_code == 429
 
     assert len(callbacks) == 1
-    assert (
-        caplog.records[-1].message
-        == "on_breach callback failed with error division by zero"
-    )
+
+    log = caplog.records[-1]
+    assert log.message == "on_breach callback failed with error division by zero"
+    assert log.levelname == "ERROR"
 
 
 def test_on_breach_callback_custom_response(extension_factory):
@@ -670,6 +705,12 @@ def test_on_breach_callback_custom_response(extension_factory):
             f"default custom response {request_limit.limit} @ {request.path}", 429
         )
 
+    def on_breach_invalid():
+        ...
+
+    def on_breach_fail(request_limit):
+        1 / 0
+
     app, limiter = extension_factory(on_breach=default_on_breach_with_response)
 
     @app.route("/")
@@ -677,29 +718,47 @@ def test_on_breach_callback_custom_response(extension_factory):
     def root():
         return "root"
 
-    @app.route("/other")
+    @app.route("/t1")
     @limiter.limit("1/second")
-    def other():
-        return "other"
+    def t1():
+        return "t1"
 
-    @app.route("/fail")
+    @app.route("/t2")
     @limiter.limit("1/second", on_breach=on_breach_with_response)
-    def fail():
-        return "fail"
+    def t2():
+        return "t2"
+
+    @app.route("/t3")
+    @limiter.limit("1/second", on_breach=on_breach_invalid)
+    def t3():
+        return "t3"
+
+    @app.route("/t4")
+    @limiter.limit("1/second", on_breach=on_breach_fail)
+    def t4():
+        return "t4"
 
     with app.test_client() as cli:
         assert cli.get("/").status_code == 200
         resp = cli.get("/")
         assert resp.status_code == 429
         assert resp.text == "default custom response 1 per 1 second @ /"
-        assert cli.get("/other").status_code == 200
-        resp = cli.get("/other")
+        assert cli.get("/t1").status_code == 200
+        resp = cli.get("/t1")
         assert resp.status_code == 429
-        assert resp.text == "default custom response 1 per 1 second @ /other"
-        assert cli.get("/fail").status_code == 200
-        resp = cli.get("/fail")
+        assert resp.text == "default custom response 1 per 1 second @ /t1"
+        assert cli.get("/t2").status_code == 200
+        resp = cli.get("/t2")
         assert resp.status_code == 429
-        assert resp.text == "custom response 1 per 1 second @ /fail"
+        assert resp.text == "custom response 1 per 1 second @ /t2"
+        resp = cli.get("/t3")
+        assert resp.status_code == 200
+        resp = cli.get("/t3")
+        assert resp.status_code == 500
+        resp = cli.get("/t4")
+        assert resp.status_code == 200
+        resp = cli.get("/t4")
+        assert resp.status_code == 500
 
 
 def test_limit_multiple_cost(extension_factory):
@@ -772,3 +831,159 @@ def test_shared_limit_multiple_cost_callable(extension_factory):
             assert 200 == cli.get("/t1").status_code
             assert 200 == cli.get("/t2").status_code
             assert 429 == cli.get("/t2").status_code
+
+
+def test_non_route_decoration_static_limits_override_defaults(extension_factory):
+    app, limiter = extension_factory(default_limits=["1/second"])
+
+    @limiter.limit("2/second")
+    def limited():
+        return "limited"
+
+    @app.route("/t1")
+    def route1():
+        return "t1"
+
+    @app.route("/t2")
+    @limiter.limit("2/second")
+    def route2():
+        return "t2"
+
+    @app.route("/t3")
+    def route3():
+        return limited()
+
+    @app.route("/t4")
+    def route4():
+        @limiter.limit("2/day", override_defaults=False)
+        def __inner():
+            return "inner"
+
+        return __inner()
+
+    @app.route("/t5/<int:param>")
+    def route5(param: int):
+        @limiter.limit("2/day", override_defaults=False)
+        def __inner1():
+            return "inner1"
+
+        @limiter.limit("3/day")
+        def __inner2():
+            return "inner2"
+
+        return __inner1() if param < 10 else __inner2()
+
+    with hiro.Timeline().freeze() as timeline:
+        with app.test_client() as cli:
+            assert 200 == cli.get("/t1").status_code
+            assert 429 == cli.get("/t1").status_code
+            for i in range(2):
+                assert 200 == cli.get("/t2").status_code
+            assert 429 == cli.get("/t2").status_code
+            for i in range(2):
+                assert 200 == cli.get("/t3").status_code, i
+            assert 429 == cli.get("/t3").status_code
+            assert 200 == cli.get("/t4").status_code
+            assert 429 == cli.get("/t4").status_code
+            timeline.forward(1)
+            assert 200 == cli.get("/t4").status_code
+            timeline.forward(1)
+            assert 429 == cli.get("/t4").status_code
+            assert 200 == cli.get("/t5/1").status_code
+            assert 429 == cli.get("/t5/1").status_code
+            timeline.forward(1)
+            assert 200 == cli.get("/t5/1").status_code
+            timeline.forward(1)
+            assert 429 == cli.get("/t5/1").status_code
+            timeline.forward(60 * 60 * 24)
+            assert 200 == cli.get("/t5/11").status_code
+            assert 200 == cli.get("/t5/11").status_code
+            assert 200 == cli.get("/t5/11").status_code
+            assert 429 == cli.get("/t5/11").status_code
+
+
+def test_non_route_decoration_static_limits(extension_factory):
+    app, limiter = extension_factory()
+
+    @limiter.limit("1/second")
+    def limited():
+        return "limited"
+
+    @app.route("/t1")
+    def route1():
+        return limited()
+
+    with hiro.Timeline().freeze():
+        with app.test_client() as cli:
+            assert 200 == cli.get("/t1").status_code
+            assert 429 == cli.get("/t1").status_code
+
+
+def test_non_route_decoration_dynamic_limits(extension_factory):
+    app, limiter = extension_factory()
+
+    def dynamic_limit_provider():
+        return "1/second"
+
+    @limiter.limit(dynamic_limit_provider)
+    def limited():
+        return "limited"
+
+    @app.route("/t1")
+    def route1():
+        return limited()
+
+    with hiro.Timeline().freeze():
+        with app.test_client() as cli:
+            assert 200 == cli.get("/t1").status_code
+            assert 429 == cli.get("/t1").status_code
+
+
+def test_non_route_decoration_multiple_sequential_limits_per_request(extension_factory):
+    app, limiter = extension_factory()
+
+    @limiter.limit("10/second")
+    def l1():
+        return "l1"
+
+    @limiter.limit("1/second")
+    def l2():
+        return "l2"
+
+    @app.route("/t1")
+    def route1():
+        return l1() + l2()
+
+    with hiro.Timeline().freeze():
+        with app.test_client() as cli:
+            assert 200 == cli.get("/t1").status_code
+            assert 429 == cli.get("/t1").status_code
+
+
+def test_inner_function_decoration(extension_factory):
+    app, limiter = extension_factory()
+
+    @app.route("/t1")
+    def route1():
+        @limiter.limit("5/second")
+        def l1():
+            return "l1"
+
+        return l1()
+
+    @app.route("/t2")
+    def route2():
+        @limiter.limit("1/second")
+        def l1():
+            return "l1"
+
+        return l1()
+
+    with hiro.Timeline().freeze():
+        with app.test_client() as cli:
+            assert 200 == cli.get("/t1").status_code
+            assert 200 == cli.get("/t2").status_code
+            for _ in range(4):
+                assert 200 == cli.get("/t1").status_code
+            assert 429 == cli.get("/t1").status_code
+            assert 429 == cli.get("/t2").status_code
